# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

"""MQTT Publisher """

import asyncio
import logging
# import copy
import json
import uuid

import paho.mqtt.client as mqtt

from foglamp.common import logger
from foglamp.plugins.north.common.common import *


__author__ = "Praveen Garg"
__copyright__ = "Copyright (c) 2020 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_LOGGER = logger.setup(__name__, level=logging.INFO)

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'MQTT (Publisher) North Plugin',
        'type': 'string',
        'default': 'mqtt',
        'readonly': 'true'
    },
    'broker': {
        'description': 'mqtt broker',
        'type': 'string',
        'default': 'localhost',  # test.mosquitto.org
        'order': '1'
    },
    'port': {
        'description': 'Port',
        'type': 'integer',
        'default': '1883',
        'order': '2'
    },
    'topic': {
        'description': 'Topic',
        'type': 'string',
        'default': 'Google/dianomic',
        'order': '3'
    },
    'keepAliveInterval': {
        'description': 'Keep alive interval',
        'type': 'integer',
        'default': '45',
        'order': '4'
    },
    "source": {
        "description": "Source of data to be sent; Can be either readings or statistics.",
        "type": "enumeration",
        "default": "readings",
        "options": ["readings", "statistics"],
        'order': '5',
        'displayName': 'Source'
    }
}


def plugin_info():

    return {
        'name': 'MQTT North Publisher',
        'version': '3.0.0-2-gb456c3f',
        'type': 'north',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Registers mqtt publisher handler

    Args:
        config: JSON configuration document for the North plugin configuration category
    Returns:
        handle: JSON object to be used in future calls to the plugin
    Raises:
    """
    # handle = copy.deepcopy(config) # FIXME: why this fails?
    # _mqtt = MqttPublisherClient(handle)
    # handle["_mqtt"] = _mqtt
    return config


async def plugin_send(data, readings_block, stream_id):
    # stream_id (REMOVE?)
    try:
        # _mqtt = data["_mqtt"]
        _mqtt = MqttPublisherClient(data)
        is_data_sent, new_last_object_id, num_sent = _mqtt.send(readings_block)
    except Exception as ex:
        _LOGGER.exception("plugin_send error: %s", ex)
    else:
        return is_data_sent, new_last_object_id, num_sent


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    Applicable on next scheduled RUN **CHECK**
    """
    # TODO:
    #     1) mqtt_client.reinitialise()
    #     2) tls_version, tls_insecure, ciphers, certificates

    # Find diff between old config and new config
    diff = utils.get_diff(handle, new_config)
    _LOGGER.info("Config changed for MQTT North plugin: [%s]", ', '.join(diff))


def plugin_shutdown(handle):
    # _mqtt = handle["_mqtt"]
    # _mqtt.stop()
    pass


class MqttPublisherClient(object):
    """ mqtt publisher class"""

    __slots__ = ['mqtt_client', 'broker', 'port', 'topic', 'keep_alive_interval']

    def __init__(self, config):
        self.mqtt_client = mqtt.Client()
        self.broker = config['broker']['value']
        self.port = int(config['port']['value'])
        self.topic = config['topic']['value']
        self.keep_alive_interval = int(config['keepAliveInterval']['value'])

    def on_connect(self, client, userdata, flags, rc):
        """ Publisher client connected to MQTT broker """
        client.connected_flag = True

    def on_disconnect(self, client, userdata, rc):
        pass

    def on_publish(self, client, userdata, mid):
        pass

    def start(self):
        # event callbacks
        self.mqtt_client.on_connect = self.on_connect
        self.mqtt_client.on_publish = self.on_publish
        self.mqtt_client.on_disconnect = self.on_disconnect

        self.mqtt_client.connect(self.broker, self.port, self.keep_alive_interval)
        _LOGGER.info("MQTT connecting... Broker: %s, Port: %s", self.broker, self.port)

        self.mqtt_client.loop_start()

    def stop(self):
        self.mqtt_client.disconnect()
        self.mqtt_client.loop_stop()

    def send(self, readings_block):
        """Send readings block from FogLAMP """

        is_data_sent = False
        last_object_id = 0
        num_sent = 0
        try:
            self.start()
            for r in readings_block:
                last_object_id = r["id"]
                msg = dict()
                msg["asset"] = r['asset_code']
                msg["readings"] = r['reading']
                msg["timestamp"] = r['user_ts']

                self.mqtt_client.publish(self.topic, json.dumps(msg))
                num_sent += 1

            is_data_sent = True
        except Exception as ex:
            _LOGGER.exception("Failed to publish, %s", str(ex))
        finally:
            self.stop()
            _LOGGER.info("MQTT publisher client connection closed. Sent %s", num_sent)

        return is_data_sent, last_object_id, num_sent
